@@ -3,6 +3,7 @@ from typing import (Callable,
                     Generic,
                     Iterable,
                     Type,
+                    Union,
                     cast)
 
 from ground.base import (Context,
@@ -13,15 +14,24 @@ from prioq.base import PriorityQueue
 from reprit.base import generate_repr
 
 from .enums import OverlapKind
-from .event import (BinaryEvent,
-                    Event,
-                    HoleyEvent,
-                    MixedEvent,
-                    NaryEvent,
-                    ShapedEvent)
+from .event import (LeftBinaryEvent,
+                    LeftHolelessEvent,
+                    LeftHoleyEvent,
+                    LeftMixedEvent,
+                    LeftNaryEvent,
+                    LeftShapedEvent,
+                    RightBinaryEvent,
+                    RightMixedEvent,
+                    RightNaryEvent,
+                    RightShapedEvent)
 from .hints import (Orienteer,
                     SegmentEndpoints)
 
+BinaryEvent = Union[LeftBinaryEvent, RightBinaryEvent]
+NaryEvent = Union[LeftNaryEvent, RightNaryEvent]
+MixedEvent = Union[LeftMixedEvent, RightMixedEvent]
+ShapedEvent = Union[LeftShapedEvent, RightShapedEvent]
+
 
 class BinaryEventsQueueKey:
     __slots__ = 'event', 'orienteer'
@@ -42,24 +52,24 @@ class BinaryEventsQueueKey:
             # different points, but same x-coordinate,
             # the event with lower y-coordinate is processed first
             return start.y < other_start.y
-        elif event.is_right_endpoint is not other_event.is_right_endpoint:
+        elif event.is_left is not other_event.is_left:
             # same start, but one is a left endpoint
             # and the other a right endpoint,
             # the right endpoint is processed first
-            return event.is_right_endpoint
+            return not event.is_left
         # same start, both events are left endpoints
         # or both are right endpoints
         else:
             other_end_orientation = self.orienteer(event.start, event.end,
                                                    other_event.end)
             # the lowest segment is processed first
-            return (other_event.from_left
+            return (other_event.from_first
                     if other_end_orientation is Orientation.COLLINEAR
                     else (other_end_orientation
                           # the lowest segment is processed first
-                          is (Orientation.CLOCKWISE
-                              if event.is_right_endpoint
-                              else Orientation.COUNTERCLOCKWISE)))
+                          is (Orientation.COUNTERCLOCKWISE
+                              if event.is_left
+                              else Orientation.CLOCKWISE)))
 
 
 class NaryEventsQueueKey:
@@ -81,22 +91,22 @@ class NaryEventsQueueKey:
             # different points, but same x-coordinate,
             # the event with lower y-coordinate is processed first
             return start.y < other_start.y
-        elif event.is_right_endpoint is not other_event.is_right_endpoint:
+        elif event.is_left is not other_event.is_left:
             # same start, but one is a left endpoint
             # and the other a right endpoint,
             # the right endpoint is processed first
-            return event.is_right_endpoint
+            return not event.is_left
         # same start, both events are left endpoints
         # or both are right endpoints
         else:
             # the lowest segment is processed first
             return (self.orienteer(event.start, event.end, other_event.end)
-                    is (Orientation.CLOCKWISE
-                        if event.is_right_endpoint
-                        else Orientation.COUNTERCLOCKWISE))
+                    is (Orientation.COUNTERCLOCKWISE
+                        if event.is_left
+                        else Orientation.CLOCKWISE))
 
 
-class LinearBinaryEventsQueue:
+class LinearEventsQueue:
     __slots__ = 'context', '_queue'
 
     def __init__(self, context: Context) -> None:
@@ -114,28 +124,21 @@ class LinearBinaryEventsQueue:
         return self._queue.key
 
     def detect_intersection(self,
-                            below_event: BinaryEvent,
-                            event: BinaryEvent) -> None:
-        below_segment_start, below_segment_end = (below_event.start,
-                                                  below_event.end)
-        segment_start, segment_end = event.start, event.end
-        relation = self.context.segments_relation(below_segment_start,
-                                                  below_segment_end,
-                                                  segment_start, segment_end)
+                            below_event: LeftBinaryEvent,
+                            event: LeftBinaryEvent) -> None:
+        relation = self.context.segments_relation(below_event, event)
         if relation is Relation.CROSS or relation is Relation.TOUCH:
             if (event.start != below_event.start
                     and event.end != below_event.end):
                 # segments do not intersect_multipolygons at endpoints
-                point = self.context.segments_intersection(
-                        below_segment_start, below_segment_end, segment_start,
-                        segment_end)
+                point = self.context.segments_intersection(below_event, event)
                 if point != below_event.start and point != below_event.end:
                     self._divide_segment(below_event, point)
                 if point != event.start and point != event.end:
                     self._divide_segment(event, point)
         elif relation is not Relation.DISJOINT:
             # segments overlap
-            if below_event.from_left is event.from_left:
+            if below_event.from_first is event.from_first:
                 raise ValueError('Segments of the same multisegment '
                                  'should not overlap.')
             starts_equal = below_event.start == event.start
@@ -148,18 +151,18 @@ class LinearBinaryEventsQueue:
             ends_equal = event.end == below_event.end
             if ends_equal:
                 end_min = end_max = None
-            elif self.key(event.complement) < self.key(below_event.complement):
-                end_min, end_max = event.complement, below_event.complement
+            elif self.key(event.opposite) < self.key(below_event.opposite):
+                end_min, end_max = event.opposite, below_event.opposite
             else:
-                end_min, end_max = below_event.complement, event.complement
+                end_min, end_max = below_event.opposite, event.opposite
             if starts_equal:
                 # both line segments are equal or share the left endpoint
                 if not ends_equal:
-                    self._divide_segment(end_max.complement, end_min.start)
+                    self._divide_segment(end_max.opposite, end_min.start)
             elif ends_equal:
                 # the line segments share the right endpoint
                 self._divide_segment(start_min, start_max.start)
-            elif start_min is end_max.complement:
+            elif start_min is end_max.opposite:
                 # one line segment includes the other one
                 self._divide_segment(start_min, end_min.start)
                 self._divide_segment(start_min, start_max.start)
@@ -172,28 +175,22 @@ class LinearBinaryEventsQueue:
         return self._queue.pop()
 
     def register(self,
-                 endpoints: Iterable[SegmentEndpoints],
-                 from_left: bool) -> None:
+                 segments_endpoints: Iterable[SegmentEndpoints],
+                 from_first: bool) -> None:
         events_queue = self._queue
-        for start, end in endpoints:
-            if start > end:
-                start, end = end, start
-            start_event = BinaryEvent(False, start, None, from_left)
-            end_event = BinaryEvent(True, end, start_event, from_left)
-            start_event.complement = end_event
-            events_queue.push(start_event)
-            events_queue.push(end_event)
+        for segment_endpoints in segments_endpoints:
+            event = LeftBinaryEvent.from_segment_endpoints(segment_endpoints,
+                                                           from_first)
+            events_queue.push(event)
+            events_queue.push(event.opposite)
 
-    def _divide_segment(self, event: BinaryEvent, point: Point) -> None:
-        left_event = BinaryEvent(False, point, event.complement,
-                                 event.from_left)
-        right_event = BinaryEvent(True, point, event, event.from_left)
-        event.complement.complement, event.complement = left_event, right_event
-        self._queue.push(left_event)
-        self._queue.push(right_event)
+    def _divide_segment(self, event: LeftBinaryEvent, point: Point) -> None:
+        tail = event.divide(point)
+        self._queue.push(tail)
+        self._queue.push(event.opposite)
 
 
-class MixedBinaryEventsQueue:
+class MixedEventsQueue:
     __slots__ = 'context', '_queue'
 
     def __init__(self, context: Context) -> None:
@@ -211,32 +208,25 @@ class MixedBinaryEventsQueue:
         return bool(self._queue)
 
     def detect_intersection(self,
-                            below_event: MixedEvent,
-                            event: MixedEvent) -> bool:
-        below_segment_start, below_segment_end = (below_event.start,
-                                                  below_event.end)
-        segment_start, segment_end = event.start, event.end
-        relation = self.context.segments_relation(below_segment_start,
-                                                  below_segment_end,
-                                                  segment_start, segment_end)
+                            below_event: LeftMixedEvent,
+                            event: LeftMixedEvent) -> bool:
+        relation = self.context.segments_relation(below_event, event)
         if relation is Relation.CROSS or relation is Relation.TOUCH:
             if (event.start != below_event.start
                     and event.end != below_event.end):
                 # segments do not intersect_multipolygons at endpoints
-                point = self.context.segments_intersection(
-                        below_segment_start, below_segment_end, segment_start,
-                        segment_end)
+                point = self.context.segments_intersection(below_event, event)
                 if point != below_event.start and point != below_event.end:
                     self._divide_segment(below_event, point)
                 if point != event.start and point != event.end:
                     self._divide_segment(event, point)
         elif relation is not Relation.DISJOINT:
             # segments overlap
-            if below_event.from_left is event.from_left:
+            if below_event.from_first is event.from_first:
                 raise ValueError('Edges of the {geometry} '
                                  'should not overlap.'
                                  .format(geometry=('multisegment'
-                                                   if event.from_left
+                                                   if event.from_first
                                                    else 'multipolygon')))
             event.is_overlap = below_event.is_overlap = True
             starts_equal = below_event.start == event.start
@@ -249,19 +239,19 @@ class MixedBinaryEventsQueue:
             ends_equal = event.end == below_event.end
             if ends_equal:
                 end_min = end_max = None
-            elif self.key(event.complement) < self.key(below_event.complement):
-                end_min, end_max = event.complement, below_event.complement
+            elif self.key(event.opposite) < self.key(below_event.opposite):
+                end_min, end_max = event.opposite, below_event.opposite
             else:
-                end_min, end_max = below_event.complement, event.complement
+                end_min, end_max = below_event.opposite, event.opposite
             if starts_equal:
                 # both line segments are equal or share the left endpoint
                 if not ends_equal:
-                    self._divide_segment(end_max.complement, end_min.start)
+                    self._divide_segment(end_max.opposite, end_min.start)
                 return True
             elif ends_equal:
                 # the line segments share the right endpoint
                 self._divide_segment(start_min, start_max.start)
-            elif start_min is end_max.complement:
+            elif start_min is end_max.opposite:
                 # one line segment includes the other one
                 self._divide_segment(start_min, end_min.start)
                 self._divide_segment(start_min, start_max.start)
@@ -275,30 +265,19 @@ class MixedBinaryEventsQueue:
         return self._queue.pop()
 
     def register(self,
-                 segments: Iterable[SegmentEndpoints],
-                 from_left: bool) -> None:
-        queue = self._queue
-        for start, end in segments:
-            interior_to_left = True
-            if start > end:
-                start, end = end, start
-                interior_to_left = False
-            start_event = MixedEvent(False, start, None, from_left,
-                                     interior_to_left)
-            end_event = MixedEvent(True, end, start_event, from_left,
-                                   interior_to_left)
-            start_event.complement = end_event
-            queue.push(start_event)
-            queue.push(end_event)
+                 segments_endpoints: Iterable[SegmentEndpoints],
+                 from_first: bool) -> None:
+        push = self._queue.push
+        for segment_endpoints in segments_endpoints:
+            event = LeftMixedEvent.from_endpoints(segment_endpoints,
+                                                  from_first)
+            push(event)
+            push(event.opposite)
 
-    def _divide_segment(self, event: MixedEvent, point: Point) -> None:
-        left_event = MixedEvent(False, point, event.complement,
-                                event.from_left, event.interior_to_left)
-        right_event = MixedEvent(True, point, event, event.from_left,
-                                 event.interior_to_left)
-        event.complement.complement, event.complement = left_event, right_event
-        self._queue.push(left_event)
-        self._queue.push(right_event)
+    def _divide_segment(self, event: LeftMixedEvent, point: Point) -> None:
+        tail = event.divide(point)
+        self._queue.push(tail)
+        self._queue.push(event.opposite)
 
 
 class NaryEventsQueue:
@@ -319,21 +298,14 @@ class NaryEventsQueue:
         return self._queue.key
 
     def detect_intersection(self,
-                            below_event: NaryEvent,
-                            event: NaryEvent) -> None:
-        below_segment_start, below_segment_end = (below_event.start,
-                                                  below_event.end)
-        segment_start, segment_end = event.start, event.end
-        relation = self.context.segments_relation(below_segment_start,
-                                                  below_segment_end,
-                                                  segment_start, segment_end)
+                            below_event: LeftNaryEvent,
+                            event: LeftNaryEvent) -> None:
+        relation = self.context.segments_relation(below_event, event)
         if relation is Relation.CROSS or relation is Relation.TOUCH:
             if (event.start != below_event.start
                     and event.end != below_event.end):
                 # segments do not intersect_multipolygons at endpoints
-                point = self.context.segments_intersection(
-                        below_segment_start, below_segment_end, segment_start,
-                        segment_end)
+                point = self.context.segments_intersection(below_event, event)
                 if point != below_event.start and point != below_event.end:
                     self._divide_segment(below_event, point)
                 if point != event.start and point != event.end:
@@ -350,18 +322,18 @@ class NaryEventsQueue:
             ends_equal = event.end == below_event.end
             if ends_equal:
                 end_min = end_max = None
-            elif self.key(event.complement) < self.key(below_event.complement):
-                end_min, end_max = event.complement, below_event.complement
+            elif self.key(event.opposite) < self.key(below_event.opposite):
+                end_min, end_max = event.opposite, below_event.opposite
             else:
-                end_min, end_max = below_event.complement, event.complement
+                end_min, end_max = below_event.opposite, event.opposite
             if starts_equal:
                 # both line segments are equal or share the left endpoint
                 if not ends_equal:
-                    self._divide_segment(end_max.complement, end_min.start)
+                    self._divide_segment(end_max.opposite, end_min.start)
             elif ends_equal:
                 # the line segments share the right endpoint
                 self._divide_segment(start_min, start_max.start)
-            elif start_min is end_max.complement:
+            elif start_min is end_max.opposite:
                 # one line segment includes the other one
                 self._divide_segment(start_min, end_min.start)
                 self._divide_segment(start_min, start_max.start)
@@ -373,29 +345,25 @@ class NaryEventsQueue:
     def pop(self) -> NaryEvent:
         return self._queue.pop()
 
-    def register(self, endpoints: Iterable[SegmentEndpoints]) -> None:
-        queue = self._queue
-        for start, end in endpoints:
-            if start > end:
-                start, end = end, start
-            start_event = NaryEvent(False, start, None)
-            end_event = NaryEvent(True, end, start_event)
-            start_event.complement = end_event
-            queue.push(start_event)
-            queue.push(end_event)
+    def register(self, segments_endpoints: Iterable[SegmentEndpoints]) -> None:
+        push = self._queue.push
+        for segment_endpoints in segments_endpoints:
+            event = LeftNaryEvent.from_segment_endpoints(segment_endpoints)
+            push(event)
+            push(event.opposite)
 
-    def _divide_segment(self, event: NaryEvent, point: Point) -> None:
-        left_event = NaryEvent(False, point, event.complement)
-        right_event = NaryEvent(True, point, event)
-        event.complement.complement, event.complement = left_event, right_event
-        self._queue.push(left_event)
-        self._queue.push(right_event)
+    def _divide_segment(self, event: LeftNaryEvent, point: Point) -> None:
+        tail = LeftNaryEvent.divide(event, point)
+        self._queue.push(tail)
+        self._queue.push(event.opposite)
 
 
-class ShapedBinaryEventsQueue(Generic[Event]):
+class ShapedEventsQueue(Generic[LeftShapedEvent]):
     __slots__ = 'context', 'event_cls', '_queue'
 
-    def __init__(self, event_cls: Type[Event], context: Context) -> None:
+    def __init__(self,
+                 event_cls: Type[LeftShapedEvent],
+                 context: Context) -> None:
         self.event_cls, self.context = event_cls, context
         self._queue = PriorityQueue(key=partial(BinaryEventsQueueKey,
                                                 context.angle_orientation))
@@ -406,30 +374,25 @@ class ShapedBinaryEventsQueue(Generic[Event]):
         return bool(self._queue)
 
     @property
-    def key(self) -> Callable[[Event], BinaryEventsQueueKey]:
+    def key(self) -> Callable[[ShapedEvent], BinaryEventsQueueKey]:
         return self._queue.key
 
-    def detect_intersection(self, below_event: Event, event: Event) -> bool:
-        below_segment_start, below_segment_end = (below_event.start,
-                                                  below_event.end)
-        segment_start, segment_end = event.start, event.end
-        relation = self.context.segments_relation(below_segment_start,
-                                                  below_segment_end,
-                                                  segment_start, segment_end)
+    def detect_intersection(self,
+                            below_event: LeftShapedEvent,
+                            event: LeftShapedEvent) -> bool:
+        relation = self.context.segments_relation(below_event, event)
         if relation is Relation.CROSS or relation is Relation.TOUCH:
             if (event.start != below_event.start
                     and event.end != below_event.end):
                 # segments do not intersect_multipolygons at endpoints
-                point = self.context.segments_intersection(
-                        below_segment_start, below_segment_end, segment_start,
-                        segment_end)
+                point = self.context.segments_intersection(below_event, event)
                 if point != below_event.start and point != below_event.end:
                     self._divide_segment(below_event, point)
                 if point != event.start and point != event.end:
                     self._divide_segment(event, point)
         elif relation is not Relation.DISJOINT:
             # segments overlap
-            if below_event.from_left is event.from_left:
+            if below_event.from_first is event.from_first:
                 raise ValueError('Edges of the same multipolygon '
                                  'should not overlap.')
             starts_equal = below_event.start == event.start
@@ -442,10 +405,10 @@ class ShapedBinaryEventsQueue(Generic[Event]):
             ends_equal = event.end == below_event.end
             if ends_equal:
                 end_min = end_max = None
-            elif self.key(event.complement) < self.key(below_event.complement):
-                end_min, end_max = event.complement, below_event.complement
+            elif self.key(event.opposite) < self.key(below_event.opposite):
+                end_min, end_max = event.opposite, below_event.opposite
             else:
-                end_min, end_max = below_event.complement, event.complement
+                end_min, end_max = below_event.opposite, event.opposite
             if starts_equal:
                 # both line segments are equal or share the left endpoint
                 below_event.overlap_kind = event.overlap_kind = (
@@ -453,12 +416,12 @@ class ShapedBinaryEventsQueue(Generic[Event]):
                     if event.interior_to_left is below_event.interior_to_left
                     else OverlapKind.DIFFERENT_ORIENTATION)
                 if not ends_equal:
-                    self._divide_segment(end_max.complement, end_min.start)
+                    self._divide_segment(end_max.opposite, end_min.start)
                 return True
             elif ends_equal:
                 # the line segments share the right endpoint
                 self._divide_segment(start_min, start_max.start)
-            elif start_min is end_max.complement:
+            elif start_min is end_max.opposite:
                 # one line segment includes the other one
                 self._divide_segment(start_min, end_min.start)
                 self._divide_segment(start_min, start_max.start)
@@ -468,40 +431,27 @@ class ShapedBinaryEventsQueue(Generic[Event]):
                 self._divide_segment(start_min, start_max.start)
         return False
 
-    def pop(self) -> Event:
+    def pop(self) -> ShapedEvent:
         return self._queue.pop()
 
     def register(self,
-                 endpoints: Iterable[SegmentEndpoints],
-                 from_left: bool) -> None:
-        event_cls = self.event_cls
-        queue = self._queue
-        for start, end in endpoints:
-            inside_on_left = True
-            if start > end:
-                start, end = end, start
-                inside_on_left = False
-            start_event = event_cls(False, start, None, from_left,
-                                    inside_on_left)
-            end_event = event_cls(True, end, start_event, from_left,
-                                  inside_on_left)
-            start_event.complement = end_event
-            queue.push(start_event)
-            queue.push(end_event)
+                 segments_endpoints: Iterable[SegmentEndpoints],
+                 from_first: bool) -> None:
+        event_cls, push = self.event_cls, self._queue.push
+        for segment_endpoints in segments_endpoints:
+            event = event_cls.from_endpoints(segment_endpoints, from_first)
+            push(event)
+            push(event.opposite)
 
-    def _divide_segment(self, event: Event, point: Point) -> None:
-        left_event = self.event_cls(False, point, event.complement,
-                                    event.from_left, event.interior_to_left)
-        right_event = self.event_cls(True, point, event, event.from_left,
-                                     event.interior_to_left)
-        event.complement.complement, event.complement = left_event, right_event
-        self._queue.push(left_event)
-        self._queue.push(right_event)
+    def _divide_segment(self, event: LeftShapedEvent, point: Point) -> None:
+        tail = event.divide(point)
+        self._queue.push(tail)
+        self._queue.push(event.opposite)
 
 
 HolelessEventsQueue = cast(Callable[[Context],
-                                    ShapedBinaryEventsQueue[ShapedEvent]],
-                           partial(ShapedBinaryEventsQueue, ShapedEvent))
+                                    ShapedEventsQueue[LeftHolelessEvent]],
+                           partial(ShapedEventsQueue, LeftHolelessEvent))
 HoleyEventsQueue = cast(Callable[[Context],
-                                 ShapedBinaryEventsQueue[HoleyEvent]],
-                        partial(ShapedBinaryEventsQueue, HoleyEvent))
+                                 ShapedEventsQueue[LeftHoleyEvent]],
+                        partial(ShapedEventsQueue, LeftHoleyEvent))
